"""Basic operations for patches."""

from __future__ import annotations

from collections.abc import Callable, Mapping, Sequence
from typing import Any, Literal

import numpy as np
import pandas as pd
from scipy.fft import next_fast_len

import dascore as dc
from dascore.constants import DEFAULT_ATTRS_TO_IGNORE, PatchType
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import CoordManager, get_coord_manager
from dascore.core.coords import get_coord
from dascore.exceptions import ParameterError, PatchBroadcastError, UnitError
from dascore.units import DimensionalityError, Quantity, Unit, get_quantity
from dascore.utils.misc import _get_nullish
from dascore.utils.models import ArrayLike
from dascore.utils.patch import (
    _merge_aligned_coords,
    _merge_models,
    align_patch_coords,
    get_dim_axis_value,
    patch_function,
)


def set_dims(self: PatchType, **kwargs: str) -> PatchType:
    """
    Set dimension to non-dimensional coordinate.

    Parameters
    ----------
    **kwargs
        A mapping indicating old_dim: new_dim where new_dim refers to
        the name of a non-dimensional coordinate which will become a
        dimensional coordinate. The old dimensional coordinate will
        become a non-dimensional coordinate.

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # add new coordinate, random numbers length of time dim
    >>> my_coord = np.random.random(patch.coord_shapes["time"])
    >>> out = (
    ...    patch.update_coords(my_coord=("time", my_coord))  # add my_coord
    ...    .set_dims(time="my_coord") # set mycoord as dim (rather than time)
    ... )
    >>> assert "my_coord" in out.dims
    """
    cm = self.coords.set_dims(**kwargs)
    return self.new(coords=cm)


def pipe(
    self: PatchType, func: Callable[[PatchType, ...], PatchType], *args, **kwargs
) -> PatchType:
    """
    Pipe the patch to a function.

    This is primarily useful for maintaining a chain of patch calls for
    a function.

    Parameters
    ----------
    func
        The function to pipe the patch. It must take a patch instance as
        the first argument followed by any number of positional or keyword
        arguments, then return a patch.
    *args
        Positional arguments that get passed to func.
    **kwargs
        Keyword arguments passed to func.
    """
    return func(self, *args, **kwargs)


def update_attrs(self: PatchType, **attrs) -> PatchType:
    """
    Update attrs and return a new Patch.

    Parameters
    ----------
    **attrs
        attrs to add/update.
    """

    def _fast_attr_update(self, attrs):
        """A fast method for just squashing the attrs and returning new patch."""
        new = self.__new__(self.__class__)
        new._data = self.data
        new._attrs = attrs
        new._coords = self.coords
        return new

    # since we update history so often, we make a fast track for it.
    new_attrs = self.attrs.model_dump(exclude_unset=True)
    new_attrs.update(attrs)
    # pop out coords so new coords has priority.
    if len(attrs) == 1 and "history" in attrs:
        return _fast_attr_update(self, PatchAttrs(**new_attrs))
    new_coords, new_attrs = self.coords.update_from_attrs(new_attrs)
    out = dict(coords=new_coords, attrs=new_attrs, dims=self.dims)
    return self.__class__(self.data, **out)


def equals(self: PatchType, other: Any, only_required_attrs=True) -> bool:
    """
    Determine if the current patch equals another.

    Parameters
    ----------
    other
        A Patch (could be equal) or some other type (not equal)
    only_required_attrs
        If True, only compare required attributes. This helps avoid issues
        with comparing histories or custom attrs of patches, for example.
    """
    # different types are not equal
    if not isinstance(other, type(self)):
        return False
    # Different coords are not equal; can pop out coords from attrs
    if not self.coords == other.coords:
        return False
    if only_required_attrs:  # only include default fields
        attrs_to_compare = set(PatchAttrs.model_fields) - {"history", "coords"}
        attrs1 = self.attrs.model_dump(include=attrs_to_compare)
        attrs2 = other.attrs.model_dump(include=attrs_to_compare)
    else:  # include all fields but coords
        attrs1 = self.attrs.model_dump(exclude=["coords"])
        attrs2 = other.attrs.model_dump(exclude=["coords"])
    if set(attrs1) != set(attrs2):  # attrs don't have same keys; not equal
        return False
    if attrs1 != attrs2:
        # see if some values are NaNs, these should be counted equal
        not_equal = {
            x
            for x in attrs1
            if attrs1[x] != attrs2[x]
            and not (pd.isnull(attrs1[x]) and pd.isnull(attrs2[x]))
        }
        if not_equal:
            return False

    return np.equal(self.data, other.data).all()


def update(
    self: PatchType,
    data: ArrayLike | np.ndarray | None = None,
    coords: None | dict[str | Sequence[str], ArrayLike] | CoordManager = None,
    dims: Sequence[str] | None = None,
    attrs: Mapping | PatchAttrs | None = None,
) -> PatchType:
    """
    Return a copy of the Patch with updated data, coords, dims, or attrs.

    Parameters
    ----------
    data
        An array-like containing data, an xarray DataArray object, or a Patch.
    coords
        The coordinates, or dimensional labels for the data. These can be
        passed in three forms:
        {coord_name: data}
        {coord_name: ((dimensions,), data)}
        {coord_name: (dimensions, data)}
    dims
        A sequence of dimension strings. The first entry corresponds to the
        first axis of data, the second to the second dimension, and so on.
    attrs
        Optional attributes (non-coordinate metadata) passed as a dict.

    Notes
    -----
    - If both coords and attrs are defined, attrs will have priority.
    """
    data = data if data is not None else self.data
    coords = coords if coords is not None else self.coords
    if dims is None:
        dims = coords.dims if isinstance(coords, CoordManager) else self.dims
    coords = get_coord_manager(coords, dims)
    if attrs:
        coords, attrs = coords.update_from_attrs(attrs)
    else:
        _attrs = dc.PatchAttrs.from_dict(attrs or self.attrs)
        attrs = _attrs.update(coords=coords, dims=coords.dims)
    return self.__class__(data=data, coords=coords, attrs=attrs, dims=coords.dims)


@patch_function()
def abs(patch: PatchType) -> PatchType:
    """
    Take the absolute value of the patch data.

    Examples
    --------
    >>> import dascore # import dascore library
    >>> pa = dascore.get_example_patch() # generate example patch
    >>> out = pa.abs() # take absolute value of generated example patch data
    """
    return patch.new(data=np.abs(patch.data))


@patch_function()
def conj(patch: PatchType) -> PatchType:
    """
    Apply the complex conjugate of the patch data.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()
    >>>
    >>> # Example 1
    >>> dft = pa.dft(None)  # multi-dim dft
    >>> conj = dft.conj()
    """
    return patch.new(data=np.conj(patch.data))


@patch_function()
def real(patch: PatchType) -> PatchType:
    """
    Return a new patch with the real part of the data array.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()
    >>> out = pa.real()
    """
    return patch.new(data=np.real(patch.data))


@patch_function()
def imag(patch: PatchType) -> PatchType:
    """
    Return a new patch with the imaginary part of the data array.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()
    >>> out = pa.imag()
    """
    return patch.new(data=np.imag(patch.data))


@patch_function()
def angle(patch: PatchType) -> PatchType:
    """
    Return a new patch with the phase angles from the data array.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()
    >>> out = pa.angle()
    """
    return patch.new(data=np.angle(patch.data))


@patch_function()
def normalize(
    self: PatchType,
    dim: str,
    norm: Literal["l1", "l2", "max", "bit"] = "l2",
) -> PatchType:
    """
    Normalize a patch along a specified dimension.

    Parameters
    ----------
    dim
        The dimension along which the normalization takes place.
    norm
        Determines the value to divide each sample by along a given axis.
        Options are:
            l1 - divide each sample by the l1 of the axis.
            l2 - divide each sample by the l2 of the axis.
            max - divide each sample by the maximum of the absolute value of the axis.
            bit - sample-by-sample normalization (-1/+1)
    """
    axis = self.dims.index(dim)
    data = self.data
    if norm in {"l1", "l2"}:
        order = int(norm[-1])
        norm_values = np.linalg.norm(self.data, axis=axis, ord=order)
    elif norm == "max":
        norm_values = np.max(data, axis=axis)
    elif norm == "bit":
        pass
    else:
        msg = (
            f"Norm value of {norm} is not supported. "
            f"Supported values are {('l1', 'l2', 'max', 'bit')}"
        )
        raise ValueError(msg)
    if norm == "bit":
        new_data = np.divide(
            data, np.abs(data), out=np.zeros_like(data), where=np.abs(data) != 0
        )
    else:
        expanded_norm = np.expand_dims(norm_values, axis=axis)
        new_data = np.divide(
            data, expanded_norm, out=np.zeros_like(data), where=expanded_norm != 0
        )
    return self.new(data=new_data)


@patch_function()
def standardize(
    self: PatchType,
    dim: str,
) -> PatchType:
    """
    Standardize data by removing the mean and scaling to unit variance.

    The standard score of a sample x is calculated as:

    z = (x - u) / s
    where u is the mean of the training samples or zero if with_mean=False,
    and s is the standard deviation of the training samples or one if with_std=False.

    Parameters
    ----------
    dim
        The dimension along which the normalization takes place.

    Examples
    --------
    ```{python}
    import dascore as dc

    patch = dc.get_example_patch()

    # standardize along the time axis
    standardized_time = patch.standardize('time')

    # standardize along the x axis
    standardized_distance = patch.standardize('distance')
    ```
    """
    axis = self.dims.index(dim)
    data = self.data
    mean = np.mean(data, axis=axis, keepdims=True)
    std = np.std(data, axis=axis, keepdims=True)
    new_data = (data - mean) / std
    return self.new(data=new_data)


@patch_function()
def apply_ufunc(
    patch: PatchType | ArrayLike,
    other: PatchType | ArrayLike,
    operator: Callable,
    *args,
    attrs_to_ignore=DEFAULT_ATTRS_TO_IGNORE,
    **kwargs,
) -> PatchType:
    """
    Apply a ufunc-type operator to a patch.

    This is used to implement a patch's operator overloading.

    Parameters
    ----------
    patch
        The patch instance.
    other
        The other object to apply the operator element-wise. Must be either a
        non-patch which is broadcastable to the shape of the patch's data, or
        a patch which has compatible coordinates. If units are provided they
        must be compatible.
    operator
        The operator. Must be numpy ufunc-like.
    *args
        Arguments to pass to the operator.
    attrs_to_ignore
        Attributes to ignore when considering if patches are compatible.
    **kwargs
        Keyword arguments to pass to the operator.

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> from dascore.proc.basic import apply_ufunc
    >>> patch = dc.get_example_patch()
    >>> # multiply the patch by 10
    >>> new = apply_ufunc(patch, 10, np.multiply)
    >>> assert np.allclose(patch.data * 10, new.data)
    >>> # add a random value to each element of patch data
    >>> noise = np.random.random(patch.shape)
    >>> new = apply_ufunc(patch, noise, np.add)
    >>> assert np.allclose(new.data, patch.data + noise)
    >>> # subtract one patch from another. Coords and attrs must be compatible
    >>> new = apply_ufunc(patch, patch, np.subtract)
    >>> assert np.allclose(new.data, 0)

    Notes
    -----
    See [numpy's ufunc docs](https://numpy.org/doc/stable/reference/ufuncs.html)
    """

    def _get_coords_attrs_from_patches(patch, other):
        """Deal with aligning two patches."""
        # Align patches so their coords are identical and data aligned.
        patch, other_patch = align_patch_coords(patch, other)
        coords = _merge_aligned_coords(patch.coords, other_patch.coords)
        # Get new attributes.
        attrs = _merge_models(
            patch.attrs,
            other_patch.attrs,
            attrs_to_ignore=attrs_to_ignore,
        )
        other = other_patch.data
        if other_units := get_quantity(other_patch.attrs.data_units):
            other = other * other_units
        return patch, other, coords, attrs

    def _ensure_array_compatible(patch, other):
        """Deal with broadcasting a patch and an array."""
        # This handles warning from quantity.
        other = other.magnitude if hasattr(other, "magnitude") else other
        other = np.asanyarray(other)
        if patch.shape == other.shape:
            return patch
        if (patch_ndims := patch.ndim) < (array_ndims := other.ndim):
            msg = f"Cannot broadcast patch/array {patch_ndims=} {array_ndims=}"
            raise PatchBroadcastError(msg)
        patch = patch.make_broadcastable_to(other.shape)
        return patch

    def _apply_op(array1, array2, operator, reversed=False):
        """Simply apply the operator, account for reversal."""
        if reversed:
            array1, array2 = array2, array1
        return operator(array1, array2, *args, **kwargs)

    def _apply_op_units(patch, other, operator, attrs, reversed=False):
        """Apply the operation handling units attached to array."""
        data_units = get_quantity(attrs.data_units)
        data = patch.data if data_units is None else patch.data * data_units
        # other is not numpy array wrapped w/ quantity, convert to quant
        if not hasattr(other, "shape"):
            other = get_quantity(other)
        try:
            new_data_w_units = _apply_op(data, other, operator, reversed=reversed)
        except DimensionalityError:
            msg = f"{operator} failed with units {data_units} and {other.units}"
            raise UnitError(msg)
        attrs = attrs.update(data_units=str(new_data_w_units.units))
        new_data = new_data_w_units.magnitude
        return new_data, attrs

    reversed = False  # flag to indicate we need to reverse data and patch
    if not isinstance(patch, dc.Patch):
        reversed = True
        patch, other = other, patch
    # Align/broadcast patch to input
    if isinstance(other, dc.Patch):
        patch, other, coords, attrs = _get_coords_attrs_from_patches(patch, other)
    else:
        patch = _ensure_array_compatible(patch, other)
        coords, attrs = patch.coords, patch.attrs
    # Apply operation
    if isinstance(other, Quantity | Unit):
        new_data, attrs = _apply_op_units(patch, other, operator, attrs, reversed)
    else:
        new_data = _apply_op(patch.data, other, operator, reversed)
    new = patch.new(data=new_data, coords=coords, attrs=attrs)
    return new


# This is left here to not break compatibility.
apply_operator = apply_ufunc


@patch_function()
def dropna(patch: PatchType, dim, how: Literal["any", "all"] = "any") -> PatchType:
    """
    Return a patch with nullish values dropped along dimension.

    Parameters
    ----------
    patch
        The patch which may contain nullish values.
    dim
        The dimension along which to drop nullish values.
    how
        "any" or "all". If "any" drop label if any null values.
        If "all" drop label if all values are nullish.

    Notes
    -----
    "nullish" is defined by `pandas.isnull`.

    Examples
    --------
    >>> import dascore as dc
    >>> # load an example patch which has some NaN values.
    >>> patch = dc.get_example_patch("patch_with_null")
    >>> # drop all time labels that have a single null value
    >>> out = patch.dropna("time", how="any")
    >>> # drop all distance labels that have all null values
    >>> out = patch.dropna("distance", how="all")
    """
    axis = patch.dims.index(dim)
    func = np.any if how == "any" else np.all
    to_drop = pd.isnull(patch.data)
    # need to iterate each non-dim axis and collapse with func
    axes = set(range(len(patch.shape))) - {axis}
    to_drop = func(to_drop, axis=tuple(axes))
    to_keep = ~to_drop
    assert len(to_keep.shape) == 1
    assert to_keep.shape[0] == patch.data.shape[axis]
    # get slices for trimming data.
    slices = [slice(None)] * len(patch.dims)
    slices[axis] = to_keep
    new_data = patch.data[tuple(slices)]
    coord = patch.get_coord(dim)
    cm = patch.coords.update(**{dim: coord[to_keep]})
    attrs = patch.attrs.update(coords={})
    return patch.new(data=new_data, coords=cm, attrs=attrs)


@patch_function()
def fillna(patch: PatchType, value) -> PatchType:
    """
    Return a patch with nullish values replaced by a value.

    Parameters
    ----------
    patch
        The patch which may contain nullish values.
    value
        The value to replace nullish values with.

    Notes
    -----
    "nullish" is defined by `pandas.isnull`.

    Examples
    --------
    >>> import dascore as dc
    >>> # load an example patch which has some NaN values.
    >>> patch = dc.get_example_patch("patch_with_null")
    >>> # Replace all occurences of NaN with 0
    >>> out = patch.fillna(0)
    >>> # Replace all occurences of NaN with 5
    >>> out = patch.fillna(5)
    """
    to_replace = pd.isnull(patch.data)

    new_data = patch.data.copy()
    new_data[to_replace] = value

    return patch.new(data=new_data)


@patch_function()
def pad(
    patch: PatchType,
    mode: Literal["constant"] = "constant",
    constant_values: Any = 0,
    expand_coords=True,
    samples=False,
    **kwargs,
) -> PatchType:
    """
    Pad the patch data along specified dimensions.

    Parameters
    ----------
    mode : str, optional
        The mode of padding, by default 'constant'.
    constant_values : scalar , optional
        A single scalar value used as the padding value across all dimensions.
        Defaults to 0.
    expand_coords : bool, optional
        Determines how coordinates are adjusted when padding is applied.
        If set to True, the coordinates will be expanded to maintain their
        order and even sampling (if evenly sampled), by extrapolating
        based on the coordinate's step size. If set to False, or coordinate
        is not evenly sampled, the new coordinates introduced by padding
        will be padded with NaN values.
    **kwargs:
        Used to specify dimension and number of elements,
        either an integer or a tuple (before, after).
        In addition, the following strings are supported:

        "fft" - pad to the next fast fft length along the given dimension by
        adding values to the end of the axis.

        "correlate" - prepare the coordinate for correlation/convolution in
        the frequency domain by pading to the next fast fft length after
        2*n - 1 where n is the current dimension length by adding values
        to the end of the axis.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # Zero pad `time` dimension with 2 patch's time unit (e.g., sec)
    >>> # zeros before and 3 zeros after
    >>> padded_patch_1 = patch.pad(time=(2, 3))
    >>> # Pad `distance` dimension with 1s 4 samples before and 4 after.
    >>> padded_patch_3 = patch.pad(distance=4, constant_values=1, samples=True)
    >>> # Get patch ready for fast fft along time dimension.
    >>> padded_fft = patch.pad(time="fft")
    """

    def _get_pad_tuple(value, samples, coord):
        """
        Get a tuple, in samples, of (pad_to_start, pad_to_end).
        """
        if value in {"fft", "correlate"}:
            target_length = len(coord) if value == "fft" else 2 * len(coord) - 1
            # Determine value so that the output dim will be a fast length.
            value = (0, next_fast_len(target_length) - len(coord))
            samples = True  # ensure padding isn't interpreted as coord units.
        elif not isinstance(value, Sequence):
            value = (value, value)
        if not samples:  # Ensure values are in samples.
            value = tuple(coord.get_sample_count(x) for x in value)
        return value

    def _get_new_coord(coord, pad_tuple, expand_coords):
        """Get the new coordinate along the expanded axis."""
        if expand_coords and coord.evenly_sampled:
            new_start = coord.min() - pad_tuple[0] * coord.step
            new_end = coord.max() + (pad_tuple[1] + 1) * coord.step
            assert coord.evenly_sampled, "expand_coords requires evenly sampled."
            new_coord = get_coord(
                start=new_start, stop=new_end, step=coord.step, units=coord.units
            )
        else:
            old_values = coord.values
            # Need to convert ints to float so NaN can be used.
            if np.issubdtype(old_values.dtype, np.integer):
                old_values = old_values.astype(np.float64)
            null_value = _get_nullish(old_values.dtype)
            added_nan_values = np.pad(
                old_values, pad_width=pad_tuple, constant_values=null_value
            )
            new_coord = coord.update(data=added_nan_values)
        return new_coord

    if isinstance(constant_values, Sequence):
        raise ParameterError("constant_values must be a scalar, not a sequence.")

    pad_width = [(0, 0)] * len(patch.shape)
    dimfo = get_dim_axis_value(patch, kwargs=kwargs, allow_multiple=True)
    new_coords = {}

    for dim, axis, value in dimfo:
        coord = patch.get_coord(dim, require_evenly_sampled=not samples)
        pad_tuple = _get_pad_tuple(value, samples, coord)
        pad_width[axis] = pad_tuple
        new_coords[dim] = _get_new_coord(coord, pad_tuple, expand_coords)

    # Pad data, update coord manager, and return.
    new_data = np.pad(patch.data, pad_width, mode=mode, constant_values=constant_values)
    new_coords = patch.coords.update(**new_coords)
    return patch.new(data=new_data, coords=new_coords)


@patch_function()
def roll(patch, samples=False, update_coord=False, **kwargs):
    """
    Roll patch array elements along a given dimension.

    Parameters
    ----------
    patch
        input patch
    samples
        if True, value indicates coordinate or value of dimension
    update_coord
        if True, updates coord based on rolled amount
    **kwargs
        specifies dimension and number of elements to roll

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # roll time dimension 5 elements
    >>> rolled_patch = patch.roll(time=5, samples=True)
    >>> # roll distance dimension 30 meters(or units of distance in patch)
    >>> rolled_patch2 = patch.roll(distance=30, samples=False)
    >>> # roll time dimension 5 elements and update coordinates
    >>> rolled_patch3 = patch.roll(time=5, samples=True, update_coord=True)
    """
    dim, axis, input_value = get_dim_axis_value(patch, kwargs=kwargs)[0]
    arr = patch.data
    coord = patch.get_coord(dim)
    value = coord.get_sample_count(input_value, samples=samples)

    roll_arr = np.roll(arr, value, axis=0)

    # update coords if True
    if update_coord:
        roll_coord_arr = np.roll(coord.values, value, axis=0)
        new_coord = coord.update(values=roll_coord_arr)
        patch = patch.update_coords(**{dim: new_coord})

    return patch.new(data=roll_arr)



"""Tests for correlate patch processing function."""

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import UnitError
from dascore.units import m
from dascore.utils.time import to_float


class TestCorrelateShift:
    """Tests for the correlation shift function."""

    @pytest.fixture(scope="class")
    def random_patch_odd(self):
        """Create a random patch with odd number of time samples."""
        patch = dc.get_example_patch("random_das", shape=(2, 11))
        return patch

    @pytest.fixture(scope="class")
    def random_patch_even(self):
        """Create a random patch with even number of time samples."""
        patch = dc.get_example_patch("random_das", shape=(2, 10))
        return patch

    def test_auto_correlation(self, random_dft_patch):
        """Perform auto correlation and undo shifting."""
        dft_conj = random_dft_patch.conj()
        dft_sq = random_dft_patch * dft_conj
        idft = dft_sq.idft()
        auto_patch = idft.correlate_shift(dim="time")
        assert np.allclose(np.imag(auto_patch.data), 0)
        assert "lag_time" in auto_patch.dims
        coord_array = auto_patch.get_array("lag_time")
        # ensure the max value happens at zero lag time.
        time_ax = auto_patch.dims.index("lag_time")
        argmax = np.argmax(random_dft_patch.data, axis=time_ax)
        assert np.all(coord_array[argmax] == dc.to_timedelta64(0))

    def test_auto_correlation_odd_coord(self, random_patch_odd):
        """Ensure correlate_shift works when dim's coord length is odd."""
        dft = random_patch_odd.dft(dim="time")
        dft_conj = dft.conj()
        dft_sq = dft * dft_conj
        idft = dft_sq.idft()
        assert isinstance(idft.correlate_shift(dim="time"), dc.Patch)

    def test_auto_correlation_even_coord(self, random_patch_even):
        """Ensure correlate_shift works when dim's coord length is even."""
        dft = random_patch_even.dft(dim="time")
        dft_conj = dft.conj()
        dft_sq = dft * dft_conj
        idft = dft_sq.idft()
        assert isinstance(idft.correlate_shift(dim="time"), dc.Patch)


class TestCorrelateInternal:
    """Tests case of intra-patch correlation function."""

    moveout_velocity = 100

    @pytest.fixture(scope="class")
    def corr_patch(self):
        """Create a patch of sin waves whose correlation can be easily checked."""
        patch = dc.get_example_patch(
            "sin_wav",
            sample_rate=100,
            frequency=range(10, 20),
            duration=5,
            channel_count=10,
        ).taper(time=0.5)
        # normalize energy so autocorrection is 1
        time_axis = patch.dims.index("time")
        data = patch.data
        norm = np.linalg.norm(data, axis=time_axis, keepdims=True)
        return patch.update(data=data / norm)

    @pytest.fixture(scope="session")
    def ricker_moveout_patch(self):
        """Create a patch with a ricker moveout."""
        patch = dc.get_example_patch(
            "ricker_moveout",
            velocity=self.moveout_velocity,
            duration=5,
        )
        return patch

    def test_basic_correlation(self, corr_patch):
        """Ensure correlation works along distance dim."""
        dist = corr_patch.get_coord("distance")[0]
        out = corr_patch.correlate(distance=dist)
        fft_ax = corr_patch.dims.index("time")
        fft_len = len(out.get_coord("lag_time"))
        argmax = np.argmax(out.data, axis=fft_ax)
        max_values = np.max(out.data, axis=fft_ax)
        # since the traces are all the same, the auto-correlation, or zero
        # lag should be maximized.
        assert np.all(np.isclose(argmax, fft_len // 2))
        # the max values should also be near 1 since the patch fixture
        # was normalized.
        assert np.allclose(max_values, 1.0, atol=0.01)
        # the sampling rates should not have changed.
        assert np.isclose(
            to_float(out.get_coord("lag_time").step),
            to_float(corr_patch.get_coord("time").step),
        )
        dstep1 = out.get_coord("distance").step
        dstep2 = corr_patch.get_coord("distance").step
        assert dstep1 == dstep2

    def test_transpose_independent(self, corr_patch):
        """The order of the dims shouldn't affect the result."""
        new_order = corr_patch.dims[::-1]
        patch1 = corr_patch
        patch2 = corr_patch.transpose(*new_order)
        corr1 = patch1.correlate(time=3, samples=True)
        corr2 = patch2.correlate(time=3, samples=True)
        assert corr1.transpose(*corr2.dims).equals(corr2)

    def test_time_lags(self, ricker_moveout_patch):
        """Ensure time lags are consistent with expected velocities."""
        corr = ricker_moveout_patch.correlate(distance=0)
        # get predicted lat times
        argmax = np.argmax(corr.data, axis=0)
        distances = corr.get_coord("distance").values
        lag_times = to_float(corr.get_coord("lag_time").values[argmax])
        # get calculated times, they should be close to lag times
        expected_times = distances / self.moveout_velocity
        assert np.allclose(lag_times.flatten(), expected_times)

    def test_units(self, random_patch):
        """Ensure units can be passed as kwarg params."""
        c_patch = random_patch.correlate(distance=10 * m)
        assert isinstance(c_patch, dc.Patch)

    def test_complex_patch(self, ricker_moveout_patch):
        """Ensure correlate works with a patch that has complex data."""
        data = ricker_moveout_patch.data
        rm_patch = ricker_moveout_patch.new(data=data + data * 1j)
        corr = rm_patch.correlate(distance=0)
        # get predicted lat times
        argmax = np.argmax(corr.data, axis=0)
        distances = corr.get_coord("distance").values
        lag_times = to_float(corr.get_coord("lag_time").values[argmax])
        # get calculated times, they should be close to lag times
        expected_times = distances / self.moveout_velocity
        assert np.allclose(lag_times.flatten(), expected_times)

    def test_correlation_freq_domain_patch(self, corr_patch):
        """
        Test correlation when the input patch is already in the frequency
        domain.
        """
        # perform FFT on the original patch to simulate frequency domain data
        fft_patch = corr_patch.dft("time")
        out = fft_patch.correlate(distance=0, samples=True)
        # The patch should still be complex.
        assert np.issubdtype(out.data.dtype, np.complexfloating)
        # Check if the shape of the output is the same as the original patch
        # plus a dimension 1 for the source.
        assert out.shape == tuple([*corr_patch.shape, 1])

    def test_correlate_decimated_patch(self, corr_patch):
        """Ensure a decimated patch can be correlated."""
        out = corr_patch.decimate(distance=2, filter_type=None).correlate(
            distance=1, samples=True
        )
        assert isinstance(out, dc.Patch)

    def test_correlate_units_raises(self, corr_patch):
        """When the patch doesn't have units an error should raise."""
        patch = corr_patch.set_units(distance=None)
        with pytest.raises(UnitError):
            patch.correlate(distance=0 * m)

    def test_correlate_units(self, corr_patch):
        """When the patch has units it should work to specify them."""
        patch = corr_patch.set_units(distance="m")
        out1 = patch.correlate(distance=1 * m)
        assert isinstance(out1, dc.Patch)
        out2 = patch.correlate(distance=np.array([1, 2]) * m)
        assert isinstance(out2, dc.Patch)

    def test_lag_deprecated(self, corr_patch):
        """Ensure the lag parameter is deprecated."""
        with pytest.warns(DeprecationWarning):
            corr_patch.correlate(time=1, lag=10, samples=True)



"""A module for loading examples."""

from __future__ import annotations

import tempfile
from collections.abc import Sequence
from contextlib import suppress
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.signal import chirp as spy_chirp

import dascore as dc
import dascore.core
from dascore.compat import random_state
from dascore.exceptions import UnknownExampleError
from dascore.utils.docs import compose_docstring
from dascore.utils.downloader import fetch
from dascore.utils.misc import iterate, register_func
from dascore.utils.patch import get_patch_names
from dascore.utils.time import to_timedelta64

EXAMPLE_PATCHES = {}
EXAMPLE_SPOOLS = {}


@register_func(EXAMPLE_PATCHES, key="random_das")
def random_patch(
    *,
    time_min="2017-09-18",
    time_step=to_timedelta64(1 / 250),
    time_array=None,
    distance_min=0,
    distance_step=1,
    dist_array=None,
    network="",
    station="",
    tag="random",
    shape=(300, 2_000),
):
    """
    Generate a random DAS Patch.

    Parameters
    ----------
    time_min
        The time the patch starts.
    time_step
        The step between time samples.
    time_array
        If not None, an array for time coordinate and`time_min` and
        `time_step` will not be used.
    distance_min
        The start of the distance coordinate.
    distance_step
        The spacing between distance samples.
    dist_array
        If not None, an array of distance values and `distance_min` and
        `distance_step` will not be used.
    network
        The network code.
    station
        The station designation.
    tag
        The patch tag
    shape
        The shape pf data array.
    """
    # get input data
    rand = np.random.RandomState(13)
    array = rand.random(shape)
    # create attrs
    t1 = np.atleast_1d(np.datetime64(time_min))[0]
    d1 = np.atleast_1d(distance_min)
    attrs = dict(
        distance_step=distance_step,
        time_step=to_timedelta64(time_step),
        category="DAS",
        time_min=t1,
        network=network,
        station=station,
        tag=tag,
        time_units="s",
        distance_units="m",
    )
    # need to pop out dim attrs if coordinates provided.
    if time_array is not None:
        attrs.pop("time_min")
        # need to keep time_step if time_array is len 1 to get coord range
        if len(time_array) > 1:
            attrs.pop("time_step")
    else:
        time_array = dascore.core.get_coord(
            data=t1 + np.arange(array.shape[1]) * attrs["time_step"],
            step=attrs["time_step"],
            units=attrs["time_units"],
        )
    if dist_array is not None:
        attrs.pop("distance_step")
    else:
        dist_array = dascore.core.get_coord(
            data=d1 + np.arange(array.shape[0]) * attrs["distance_step"],
            step=attrs["distance_step"],
            units=attrs["distance_units"],
        )
    coords = dict(distance=dist_array, time=time_array)
    # assemble and output.
    out = dict(data=array, coords=coords, attrs=attrs, dims=("distance", "time"))
    patch = dc.Patch(**out)
    return patch


@register_func(EXAMPLE_PATCHES, key="patch_with_null")
def patch_with_null(**kwargs):
    """
    A patch which has nullish values.

    Parameters
    ----------
    **kwargs
        Parameters passed to [`random_patch`](`dascore.examples.random_patch`).
    """
    patch = random_patch(**kwargs)
    data = np.array(patch.data)
    data[data > 0.9] = np.nan
    # also set the first row and column to NaN
    data[:, 0] = np.nan
    data[0, :] = np.nan
    return patch.new(data=data)


@register_func(EXAMPLE_PATCHES, key="random_patch_with_lat_lon")
def random_patch_lat_lon(**kwargs):
    """
    Create a patch with latitude/longitude coords on distance dimension.

    Parameters
    ----------
    **kwargs
        Parameters passed to [`random_patch`](`dascore.examples.random_patch`).
    """
    patch = random_patch(**kwargs)
    dist = patch.coords.get_array("distance")
    lat = np.arange(0, len(dist)) * 0.001 - 109.857952
    lon = np.arange(0, len(dist)) * 0.001 + 41.544654
    # add a single coord
    out = patch.update_coords(latitude=("distance", lat), longitude=("distance", lon))
    return out


@register_func(EXAMPLE_PATCHES, key="random_patch_with_xyz")
def random_patch_xyz(**kwargs):
    """
    Create a patch with x, y, and z coords on distance dimension.

    Parameters
    ----------
    **kwargs
        Parameters passed to [`random_patch`](`dascore.examples.random_patch`).
    """
    patch = random_patch(**kwargs)
    dist = patch.coords.get_array("distance")
    x = np.arange(0, len(dist)) * 5
    y = np.arange(0, len(dist)) * 5
    z = np.zeros_like(dist)
    # add a single coord
    out = patch.update_coords(x=("distance", x), y=("distance", y), z=("distance", z))
    return out


@register_func(EXAMPLE_PATCHES, key="wacky_dim_coords_patch")
def wacky_dim_coord_patch():
    """
    A patch with one Monotonic and one Array coord.
    """
    shape = (100, 1_000)
    # distance is neither monotonic nor evenly sampled.
    dist_ar = random_state.random(100) + np.arange(100) * 0.3
    # time is monotonic, not evenly sampled.
    time_ar = dc.to_datetime64(np.cumsum(random_state.random(1_000)))
    patch = random_patch(shape=shape, dist_array=dist_ar, time_array=time_ar)
    # check attrs
    attrs = patch.attrs
    assert pd.isnull(attrs.coords["time"].step)
    assert pd.isnull(attrs.coords["time"].step)
    return patch


@register_func(EXAMPLE_PATCHES, key="sin_wav")
def sin_wave_patch(
    sample_rate=44100,
    frequency: Sequence[float] | float = 100.0,
    time_min="2020-01-01",
    channel_count=3,
    duration=1,
    amplitude: Sequence[float] | float = 10.0,
):
    """
    A Patch composed of sine waves.

    Parameters
    ----------
    sample_rate
        The sample rate in Hz.
    frequency
        The frequency of the sin wave. If a sequence is provided multiple
        sine waves will be generated at each frequency.
    time_min
        The start time in the metadata.
    channel_count
        The number of  distance channels to include.
    duration
        Signal duration in seconds.
    amplitude
        The amplitude of the sin wave. If a sequence is provided it represents
        the amplitude of each frequency.
    """
    t_array = np.linspace(0.0, duration, int(sample_rate * duration))
    # Get time and distance coords
    distance = np.arange(1, channel_count + 1, 1)
    time = to_timedelta64(t_array) + np.datetime64(time_min)
    freqs = np.atleast_1d(frequency)
    amps = np.broadcast_to(np.atleast_1d(amplitude), shape=freqs.shape)
    # init empty data and add frequencies.
    data = np.zeros((len(time), len(distance)))
    for amp, freq in zip(amps, freqs):
        sin_data = amp * np.sin(2.0 * np.pi * freq * t_array)
        data += sin_data[..., np.newaxis]
    patch = dc.Patch(
        data=data,
        coords={"time": time, "distance": distance},
        dims=("time", "distance"),
    )
    return patch


@register_func(EXAMPLE_PATCHES, key="chirp")
def chirp(
    sample_rate=150,
    time_min="2020-01-01",
    channel_count: int = 1,
    duration: float = 10.0,
    f0: float = 5.0,
    t1: float | None = None,
    f1: float = 25.0,
    method="linear",
    phi: float = 0.0,
    **kwargs,
):
    """
    Create a patch from a chirp signal.

    Simply uses scipy.signal.chirp under the hood.

    Parameters
    ----------
    sample_rate
        The sample rate in Hz.
    time_min
        The start time in the metadata.
    channel_count
        The number of  distance channels to include.
    duration
        The duration, in seconds, of the signal.
    f0
        The frequency of the chirp at the start of the signal.
    f1
        The frequency of the chirp at the end of the signal.
    t1
        The time (relative from signal start) corresponding to f1. If None,
        use the end of the signal.
    method
        The kind of the frequency sweep. See scipy.signal.chirp for
        more details.
    phi
        Phase offset in degrees.
    **kwargs
        Passed directly to scipy.signal.chirp.
    """
    t_array = np.linspace(0.0, duration, int(sample_rate * duration))
    t1 = t1 if t1 is not None else np.max(t_array)
    array = spy_chirp(t_array, f0=f0, t1=t1, f1=f1, method=method, phi=phi, **kwargs)
    # Get time and distance coords
    distance = np.arange(1, channel_count + 1, 1)
    time = to_timedelta64(t_array) + np.datetime64(time_min)
    data = np.array([array for _ in range(len(distance))])
    patch = dc.Patch(
        data=data,
        coords={"time": time, "distance": distance},
        dims=("distance", "time"),
    )
    return patch


@register_func(EXAMPLE_PATCHES, key="example_event_1")
def example_event_1():
    """
    An induced event recorded on a borehole fiber  from @stanvek2022fracture.
    """
    path = fetch("example_dasdae_event_1.h5")
    return dc.spool(path)[0]


@register_func(EXAMPLE_PATCHES, key="example_event_2")
def example_event_2():
    """
    [`example_event_1`](`dascore.examples.example_event_1`) with pre-processing.
    """
    path = fetch("example_dasdae_event_1.h5")
    patch = dc.spool(path)[0].update_attrs(data_type="strain_rate")
    # We convert time to relative time in seconds to match the figure in
    # the publication.
    delta_time = patch.coords.get_array("time") - patch.coords.min("time")
    out = (
        patch.update_coords(time=delta_time / np.timedelta64(1, "s"))
        .set_units("strain/s", distance="m", time="s")
        .taper(time=0.05)
        .pass_filter(time=(..., 300))
    )
    return out


@register_func(EXAMPLE_PATCHES, key="deformation_rate_event_1")
def deformation_rate_event_1():
    """
    An event recorded in an underground mine by a Terra15 unit.
    """
    path = fetch("deformation_rate_event_1.hdf5")
    return dc.spool(path)[0]


@register_func(EXAMPLE_PATCHES, key="forge_dss")
def forge_dss():
    """
    A DSS file from the Forge dataset collected by Neubrex.

    https://gdr.openei.org/submissions/1565
    """
    path = fetch("neubrex_dss_forge.h5")
    return dc.spool(path)[0]


@register_func(EXAMPLE_PATCHES, key="forge_dts")
def forge_dts():
    """
    A DTS file from the Forge dataset collected by Neubrex.

    https://gdr.openei.org/submissions/1565
    """
    path = fetch("neubrex_dts_forge.h5")
    return dc.spool(path)[0]


@register_func(EXAMPLE_PATCHES, key="ricker_moveout")
def ricker_moveout(
    frequency=15,
    peak_time=0.25,
    duration=1.5,
    time_step=0.002,
    distance_step=10,
    channel_count=10,
    source_channel=0,
    velocity=100,
):
    """
    A patch of a ricker wavelet with some apparent velocity.

    Parameters
    ----------
    frequency
        The center frequency of the wavelet in Hz.
    peak_time
        The peak time of the first ricker wavelet in seconds.
    duration
        The total duration of the time coordinate in seconds.
    time_step
        The time dimension time step.
    distance_step
        The distance dimension sampling interval.
    channel_count
        The total number of channels (number of distance).
    source_channel
        The index of the source.
    velocity
        The apparent velocity in m/s.

    Notes
    -----
    Based on https://github.com/lijunzh/ricker/.
    """

    def _ricker(time, delay):
        # shift time vector to account for different peak times.
        new_time = time - delay
        f = frequency
        # get amplitude and exp term of ricker
        const = 1 - 2 * np.pi**2 * f**2 * new_time**2
        exp = np.exp(-(np.pi**2) * f**2 * new_time**2)
        return const * exp

    time = np.arange(0, duration + time_step, time_step)
    distance = np.arange(channel_count) * distance_step
    assert source_channel < len(distance)
    source_distance = distance[source_channel]
    data = np.zeros((len(time), len(distance)))
    # iterate each distance channel and update data
    for ind, dist in enumerate(distance):
        dist_to_source = np.abs(dist - source_distance)
        time_delay = peak_time + (dist_to_source / velocity)
        data[:, ind] = _ricker(time, time_delay)

    coords = {"time": to_timedelta64(time), "distance": distance}
    dims = ("time", "distance")
    return dc.Patch(data=data, coords=coords, dims=dims)


@register_func(EXAMPLE_PATCHES, key="delta_patch")
def delta_patch(
    dim="time",
    shape=(10, 200),
    time_min="2020-01-01",
    time_step=1 / 250,
    distance_min=0,
    distance_step=1,
    patch=None,
):
    """
    Create a delta function patch (zeros everywhere except for
    a unit value at the center) along the specified dimension.
    The returned delta patch has single coordinate(s) along the
    other dimensions.

    Parameters
    ----------
    dim : str
        The dimension at the center of which to place the unit value.
        Typically ``"time"`` or ``"distance"``.
    shape : tuple of int
        The shape of the data as (distance, time). Defaults to (10, 200).
        This is used only if no existing ``patch`` is provided.
    time_min : str or datetime64
        The start time of the patch.
    time_step : float
        The time step in seconds between samples.
    distance_min : float
        The minimum distance coordinate.
    distance_step : float
        The distance step in meters between samples.
    patch : dascore.Patch
        If provided, creates the delta patch based on this existing patch.
        Default is None.
    """
    if patch is None:
        if dim not in ["time", "distance"]:
            raise ValueError(
                "In case no patch is provided, the delta patch will be "
                "a 2D patch with 'time' and 'distance' dimensions."
            )

        dims = ("distance", "time")
        dist_len, time_len = shape

        # Create coordinates
        time_step_td = to_timedelta64(time_step)
        t0 = np.datetime64(time_min)
        time_coord = dascore.core.get_coord(
            data=t0 + np.arange(time_len) * time_step_td, step=time_step_td, units="s"
        )
        dist_coord = dascore.core.get_coord(
            data=distance_min + np.arange(dist_len) * distance_step,
            step=distance_step,
            units="m",
        )

        coords = {"distance": dist_coord, "time": time_coord}
        attrs = dict(
            time_min=t0,
            time_step=time_step_td,
            distance_min=distance_min,
            distance_step=distance_step,
            category="DAS",
            network="",
            station="",
            tag="delta",
            time_units="s",
            distance_units="m",
        )

        # Depending on the selected dimension, place a line of ones at the midpoint
        used_dims = tuple(iterate(dim))
        unused_dims = set(dims) - set(used_dims)

        # Get data with ones centered on selected dimensions.
        index = tuple(
            shape[dims.index(dimension)] // 2 if dimension in used_dims else 0
            for dimension in dims
        )
        data = np.zeros((dist_len, time_len))
        data[index] = 1
        delta_patch = dc.Patch(data=data, coords=coords, dims=dims, attrs=attrs)
        return delta_patch.select(**{x: 0 for x in unused_dims}, samples=True)
    else:
        used_dims = tuple(iterate(dim))
        unused_dims = set(patch.dims) - set(used_dims)
        patch = patch.select(**{x: 0 for x in unused_dims}, samples=True)

        # Get data with ones centered on selected dimensions.
        shape = patch.shape
        index = tuple(
            shape[patch.dims.index(dimension)] // 2 if dimension in used_dims else 0
            for dimension in patch.dims
        )
        data = np.zeros_like(patch.data)
        data[index] = 1
        return patch.update(data=data)


@register_func(EXAMPLE_PATCHES, key="dispersion_event")
def dispersion_event():
    """
    A synthetic shot record that exhibits dispersion.
    """
    path = fetch("dispersion_event.h5")
    return dc.spool(path)[0]


@register_func(EXAMPLE_SPOOLS, key="random_das")
def random_spool(time_gap=0, length=3, time_min=np.datetime64("2020-01-03"), **kwargs):
    """
    Several random patches in the spool.

    Parameters
    ----------
    time_gap
        The difference in time between each patch. Use a negative
        number to create overlap.
    length
        The number of patches to generate.
    time_min
        The start time of the first patch. Subsequent patches have start times
        after the end time of the previous patch, plus the time_gap.
    **kwargs
        Passed to the [_random_patch](`dascore.examples.random_patch`) function.
    """
    out = []
    for _ in range(length):
        patch = random_patch(time_min=time_min, **kwargs)
        out.append(patch)
        diff = to_timedelta64(time_gap) + patch.attrs.coords["time"].step
        time_min = patch.attrs["time_max"] + diff
    return dc.spool(out)


@register_func(EXAMPLE_SPOOLS, key="random_directory_das")
def random_directory_spool(path=None, **kwargs):
    """
    Create a random spool, then save to specified path.

    Parameters
    ----------
    path
        If provided, the path to save the directory spool. If None, use
        a temporary path.

    kwargs are passed to [`random_spool`](`dascore.examples.random_spool`)
    """
    spool = random_spool(**kwargs)
    path = spool_to_directory(spool, path)
    return dc.spool(path)


@register_func(EXAMPLE_SPOOLS, key="diverse_das")
def diverse_spool():
    """
    A spool with a diverse set of patch metadata for testing.

    There are various gaps, tags, station names, etc.
    """
    spool_no_gaps = random_spool()
    spool_no_gaps_different_network = random_spool(network="das2")
    spool_big_gaps = random_spool(time_gap=np.timedelta64(1, "s"), station="big_gaps")
    spool_overlaps = random_spool(
        time_gap=-np.timedelta64(10, "ms"), station="overlaps"
    )
    time_step = spool_big_gaps[0].attrs.coords["time"].step
    dt = to_timedelta64(time_step / np.timedelta64(1, "s"))
    spool_small_gaps = random_spool(time_gap=dt, station="smallg")
    spool_way_late = random_spool(
        length=1, time_min=np.datetime64("2030-01-01"), station="wayout"
    )
    spool_new_tag = random_spool(tag="some_tag", length=1)
    spool_way_early = random_spool(
        length=1, time_min=np.datetime64("1989-05-04"), station="wayout"
    )

    all_spools = [
        spool_no_gaps,
        spool_no_gaps_different_network,
        spool_big_gaps,
        spool_overlaps,
        spool_small_gaps,
        spool_way_late,
        spool_new_tag,
        spool_way_early,
    ]

    return dc.spool([y for x in all_spools for y in x])


def spool_to_directory(spool, path=None, file_format="DASDAE", extention="hdf5"):
    """
    Write out each patch in a spool to a directory.

    Parameters
    ----------
    spool
        The spool to save to
    path
        The path to the directory, if None, create tempdir.
    file_format
        The file format for the saved files.
    """
    if path is None:
        path = Path(tempfile.mkdtemp())
        assert path.exists()
    for patch in spool:
        name = get_patch_names(patch).iloc[0]
        out_path = path / (f"{name}.{extention}")
        patch.io.write(out_path, file_format=file_format)
    return path


@compose_docstring(examples=", ".join(list(EXAMPLE_PATCHES)))
def get_example_patch(example_name="random_das", **kwargs) -> dc.Patch:
    """
    Load an example Patch.

    Options are:
    ```{python}
    #| echo: false
    #| output: asis
    from dascore.examples import EXAMPLE_PATCHES

    from dascore.utils.docs import objs_to_doc_df

    df = objs_to_doc_df(EXAMPLE_PATCHES)
    print(df.to_markdown(index=False, stralign="center"))
    ```

    Using an entry from the data_registry file is also supported.
    If multiple patches are contained in the specified file, only the
    first is returned. Data registry files are:
    ```{python}
    #| echo: false
    #| output: asis
    from dascore.utils.downloader import get_registry_df
    print(get_registry_df()[['name']].to_markdown(index=False, stralign="center"))
    ```

    Parameters
    ----------
    example_name
        The name of the example to load. Options are listed above.
    **kwargs
        Passed to the corresponding functions to generate data.

    Raises
    ------
        (`UnknownExampleError`)['dascore.examples.UnknownExampleError`] if
        unregistered patch is requested.
    """
    if example_name not in EXAMPLE_PATCHES:
        # Allow the example name to be a data registry entry.
        with suppress(ValueError):
            return dc.spool(fetch(example_name))[0]
        msg = (
            f"No example patch registered with name {example_name} "
            f"Registered example patches are {list(EXAMPLE_PATCHES)}"
        )
        raise UnknownExampleError(msg)
    return EXAMPLE_PATCHES[example_name](**kwargs)


@compose_docstring(examples=", ".join(list(EXAMPLE_SPOOLS)))
def get_example_spool(example_name="random_das", **kwargs) -> dc.BaseSpool:
    """
    Load an example Spool.

    Supported example spools are:
    ```{python}
    #| echo: false
    #| output: asis
    from dascore.examples import EXAMPLE_SPOOLS

    from dascore.utils.docs import objs_to_doc_df

    df = objs_to_doc_df(EXAMPLE_SPOOLS)
    print(df.to_markdown(index=False, stralign="center"))
    ```

    Using an entry from the data_registry file is also supported.
    These include:
    ```{python}
    #| echo: false
    #| output: asis
    from dascore.utils.downloader import get_registry_df
    print(get_registry_df()[['name']].to_markdown(index=False, stralign="center"))
    ```

    Parameters
    ----------
    example_name
        The name of the example to load. Options are:
    **kwargs
        Passed to the corresponding functions to generate data.

    Raises
    ------
    (`UnknownExampleError`)['dascore.examples.UnknownExampleError`] if
        unregistered patch is requested.
    """
    if example_name not in EXAMPLE_SPOOLS:
        # Allow the example spool to be a data registry file.
        with suppress(ValueError):
            return dc.spool(fetch(example_name))
        msg = (
            f"No example spool registered with name {example_name} "
            f"Registered example spools are {list(EXAMPLE_SPOOLS)}"
        )
        raise UnknownExampleError(msg)
    return EXAMPLE_SPOOLS[example_name](**kwargs)



"""Simple interface for progress markers."""

from __future__ import annotations

from collections.abc import Generator, Sized
from contextlib import suppress

import rich.progress as prog

import dascore as dc
from dascore.compat import Progress
from dascore.constants import PROGRESS_LEVELS


def get_progress_instance(progress: PROGRESS_LEVELS | Progress = "standard"):
    """
    Get the Rich progress bar instance based on complexity level.
    """
    # If a progress class is passed in, just use it.
    if isinstance(progress, Progress):
        return progress
    kwargs = {}
    progress_list = [
        prog.SpinnerColumn(),
        prog.TextColumn("[progress.description]{task.description}"),
        prog.BarColumn(bar_width=30),
        prog.TaskProgressColumn(),
        prog.TimeRemainingColumn(),
        prog.TimeElapsedColumn(),
        prog.MofNCompleteColumn(),
    ]
    if progress == "basic":
        # set the refresh rate very low and eliminate the spinner
        kwargs["refresh_per_second"] = 0.25
        progress_list = progress_list[1:]
    return Progress(*progress_list, **kwargs)


def track(
    sequence: Sized | Generator,
    description: str,
    progress: PROGRESS_LEVELS | Progress = "standard",
    length: int | None = None,
    min_length: int = 1,
):
    """
    A simple iterator for tracking updates.

    Parameters
    ----------
    sequence
        A sequence or generator to trace the iteration over.
    description
        A string describing the operation
    progress
        options are
            None- disable progress bar,
            "basic" reduced refresh rate,
            "standard" - the normal progress bar
        can also accept a subclass of rich.progress.Progress.
    min_length
        The minimum length to emmit a progress bar.
    """
    # In the case of a generator we need to make sure this just exists
    guess_len = length if length is not None else 0
    with suppress(TypeError, ValueError):
        length = len(sequence) if not guess_len else guess_len
    if length < min_length:
        length = 0
    # This is a dirty hack to allow debugging while running tests.
    # Otherwise, pdb doesn't work in any tracking scope.
    # See: https://github.com/Textualize/rich/issues/1053
    if dc._debug or not length or progress is None:
        yield from sequence
        return
    update = 1.0 if isinstance(progress, str) and progress == "standard" else 5.0
    progress = get_progress_instance(progress)
    with progress:
        yield from progress.track(
            sequence,
            total=length or len(sequence),
            description=description,
            update_period=update,
        )



"""Utilities for OptoDAS."""

from __future__ import annotations

import dascore as dc
import dascore.core
from dascore.core.coords import get_coord
from dascore.utils.hdf5 import unpack_scalar_h5_dataset
from dascore.utils.misc import unbyte

# --- Getting format/version


def _get_opto_das_version_str(hdf_fi) -> str:
    """Return the version string for OptoDAS file."""
    # define a few root attrs that act as a "fingerprint"
    expected_attrs = (
        "acqSpec",
        "header",
        "cableSpec",
        "data",
        "fileVersion",
    )
    if not all([x in hdf_fi for x in expected_attrs]):
        return ""
    version_str = str(unbyte(hdf_fi["fileVersion"][()]))
    return version_str


def _get_coord_manager(fi):
    """Get the distance ranges and spacing."""
    header = fi["header"]
    dims = tuple(unbyte(x) for x in header["dimensionNames"])
    units = tuple(unbyte(x) for x in header["dimensionUnits"])
    coords = {}
    for index, (dim, unit) in enumerate(zip(dims, units)):
        crange = header["dimensionRanges"][f"dimension{index}"]
        step = unpack_scalar_h5_dataset(crange["unitScale"])

        # Special case for time.
        if dim == "time":
            step = dc.to_timedelta64(step)
            t1 = dc.to_datetime64(unpack_scalar_h5_dataset(header["time"]))
            start = t1 + unpack_scalar_h5_dataset(crange["min"]) * step
            stop = t1 + (unpack_scalar_h5_dataset(crange["max"]) + 1) * step
            coord = get_coord(min=start, max=stop, step=step, units=unit)
        else:  # and distance
            # The channels are ints so we multiply by step to get distance.
            distance = fi["/header/channels"][:] * step
            coord = get_coord(values=distance)
        coords[dim] = coord
    out = dascore.core.get_coord_manager(coords=coords, dims=dims)
    return out


def _get_attr_dict(header):
    """Map header info to DAS attrs."""
    attr_map = {
        "gaugeLength": "gauge_length",
        "unit": "data_units",
        "instrument": "instrument_id",
        "experiment": "acquisition_id",
    }
    out = {"data_category": "DAS"}
    for head_name, attr_name in attr_map.items():
        value = header[head_name]
        if hasattr(value, "shape"):
            value = unpack_scalar_h5_dataset(value)
        out[attr_name] = unbyte(value)
    return out


def _get_opto_das_attrs(fi) -> dict:
    """Scan a OptoDAS file, return metadata."""
    cm = _get_coord_manager(fi)
    attrs = _get_attr_dict(fi["header"])
    attrs["coords"] = cm
    return attrs


def _read_opto_das(fi, distance=None, time=None, attr_cls=dc.PatchAttrs):
    """Read the OptoDAS values into a patch."""
    attrs = _get_opto_das_attrs(fi)
    data_node = fi["data"]
    coords = attrs.pop("coords")
    cm, data = coords.select(array=data_node, distance=distance, time=time)
    attrs["coords"] = cm.to_summary_dict()
    attrs["dims"] = cm.dims
    return [dc.Patch(data=data, coords=cm, attrs=attr_cls(**attrs))]



"""Simple script for downloading external files."""

from __future__ import annotations

from functools import cache
from importlib.resources import files
from pathlib import Path

import pandas as pd
import pooch

from dascore.constants import DATA_VERSION

REGISTRY_PATH = Path(files("dascore").joinpath("data_registry.txt"))

# Create a pooch for fetching data files
fetcher = pooch.create(
    path=pooch.os_cache("dascore"),
    base_url="https://github.com/d-chambers/dascore",
    version=DATA_VERSION,
    version_dev="master",
    env="DFS_DATA_DIR",
)
fetcher.load_registry(REGISTRY_PATH)


@cache
def get_registry_df() -> pd.DataFrame:
    """Returns a dataframe of all files in the data registry."""
    names = (
        "name",
        "hash",
        "url",
    )
    df = pd.read_csv(REGISTRY_PATH, sep=r"\s+", skiprows=1, names=names)
    return df


@cache
def fetch(name: Path | str, **kwargs) -> Path:
    """
    Fetch a data file from the registry.

    Parameters
    ----------
    name
        The name of the file to fetch. Must be in the data registry or a
        path which exists.
    kwargs
        Left for compatibility reasons.

    Returns
    -------
    A path to the downloaded file.
    """
    if (existing_path := Path(name)).exists():
        return existing_path
    return Path(fetcher.fetch(name, **kwargs))
